#    Munakello - A simple Pomodoro timer
#    Copyright (C) 2016 Timo Kankare 
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""A simple Pomodoro timer"""

import argparse
import asyncio
from datetime import datetime, timedelta
import functools
import signal

from . import about

def beep():
    """Emit a beep sound."""
    print('\a', end='', flush=True)


def round_time(time, precision):
    """Round time to specified precision."""
    return precision * ((time + precision / 2) // precision)


def calculate_next_beep_time(remaining_time, end_time):
    """Calculate time for the next sound."""
    time_scales = [
        timedelta(seconds=60),
        timedelta(seconds=10),
        timedelta(seconds=3),
        timedelta(seconds=2),
        timedelta(seconds=1),
    ]
    for nextstep in time_scales:
        if remaining_time > nextstep:
            return end_time - nextstep
    else:
        return end_time


async def report_beep(start_time, end_time):
    """*coroutine* Make beep sounds at proper times."""
    try:
        while True:
            now = datetime.now()
            remaining_time = end_time - now
            if remaining_time.total_seconds() <= 0:
                break
            next_time = calculate_next_beep_time(remaining_time, end_time)
            await asyncio.sleep((next_time - now).total_seconds())
            beep()
    except asyncio.CancelledError:
        pass


def select_message(remaining_time):
    """Select message for the given time."""
    time_scales = [
        (timedelta(seconds=59), '{minutes} minutes remaining'),
        (timedelta(seconds=0), '{seconds} seconds remaining'),
    ]
    for (limit, message) in time_scales:
        if remaining_time > limit:
            return message
    return '{time} remaining'


def format_time(remaining_time):
    """Format time message."""
    message = select_message(remaining_time)
    hours = round(remaining_time.total_seconds() / (60*60))
    minutes = round(remaining_time.total_seconds() / 60)
    seconds = round(remaining_time.total_seconds())
    return message.format(hours=hours, minutes=minutes, seconds=seconds, time=remaining_time)


def calculate_next_report_time(remaining_time, end_time):
    """Calculate next time to report remaining time."""
    time_scales = [
        (timedelta(seconds=60), timedelta(seconds=60)),
        (timedelta(seconds=10), timedelta(seconds=10)),
        (timedelta(seconds=0),  timedelta(seconds=1)),
    ]
    for (limit, nextstep) in time_scales:
        t = round_time(remaining_time - nextstep, nextstep)
        if t >= limit:
            return end_time - t
    return end_time


async def report_time(start_time, end_time):
    """*coroutine* Report remaining time at suitable moments."""
    try:
        while True:
            now = datetime.now()
            remaining_time = end_time - now
            if remaining_time.total_seconds() <= 0:
                break
            print('\r' + format_time(remaining_time), end='    ', flush=True)
            next_time = calculate_next_report_time(remaining_time, end_time)
            await asyncio.sleep((next_time - now).total_seconds())
    except asyncio.CancelledError:
        pass
    finally:
        print()


async def wait_delay(delay, reporters):
    """*coroutine* Start reporters and wait the given time."""
    start_time = datetime.now()
    end_time = start_time + timedelta(seconds=delay)

    # start reporting tasks
    reporter_tasks = [asyncio.ensure_future(reporter(start_time, end_time)) for reporter in reporters]
    pending = reporter_tasks

    try:
        # wait end of delay
        await asyncio.sleep(delay)
        # wait tasks to end
        (done, pending) = await asyncio.wait(reporter_tasks, timeout=1)
    finally:
        # cancel remaining tasks
        for task in pending:
            task.cancel()


async def work(delay, start_message, stop_message, reporters):
    """*coroutine* Run work or break."""
    print(start_message)
    await wait_delay(delay, reporters)
    print(stop_message)


async def run_pomodoro(reporters, work_length, break_length):
    """*coroutine* Run one pomodoro."""
    try:
        await work(work_length * 60, 'Start working...', 'Stop working!', reporters)
        await work(break_length * 60, 'Start break...', 'Stop break.', reporters)
    except asyncio.CancelledError:
        print('Cancelled.')


def cancel_tasks(tasks):
    """Cancel given tasks."""
    for task in tasks:
        task.cancel()


def main():
    """Main function.

    Handle command line arguments, run command loop.
    """

    print("{name} {version} - {description}.".format(
        name=about.name.capitalize(), 
        version=about.version, 
        description=about.description)
    )

    parser = argparse.ArgumentParser(description='Counts time for working period and break period.')
    parser.add_argument('--no-sound', dest='sound_reporters', 
                        action='store_const',
                        const=[], default=[report_beep],
                        help='prevent sound')
    parser.add_argument('--work-length', dest='work_length', 
                        action='store', type=float,
                        default=25, metavar='minutes',
                        help='length of work period in minutes')
    parser.add_argument('--break-length', dest='break_length', 
                        action='store', type=float,
                        default=5, metavar='minutes',
                        help='length of break period in minutes')
    args = parser.parse_args()

    reporters = [report_time] + args.sound_reporters
    loop = asyncio.get_event_loop()

    tasks = [ 
        asyncio.ensure_future(run_pomodoro(reporters, 
                                           args.work_length, 
                                           args.break_length)) 
    ]

    try:
        for signame in ('SIGINT', 'SIGTERM'):
            loop.add_signal_handler(getattr(signal, signame),
                                    functools.partial(cancel_tasks, tasks))
    except:
        # ignore errors
        pass 

    try:
        loop.run_until_complete(asyncio.wait(tasks))
    finally:
        loop.close()


if __name__ == '__main__':
    main()

